﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Linq;

  using Common;
  using Configuration;
  using Core;
  using HeuristicLab.Data;
  using HeuristicLab.Encodings.IntegerVectorEncoding;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Analyzer;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Individual;
  using HeuristicLab.Problems.ProgramSynthesis.Push.SolutionCreator;
  using Optimization;
  using Persistence.Default.CompositeSerializers.Storable;

  [StorableClass]
  public abstract class IntegerVectorPushProblem : PushProblemBase<IntegerVectorEncoding> {
    protected IntegerVectorPushProblem(IPushEvaluator evaluator) : base(evaluator) {
      InitEvents();
      InitEncoding();
      InitOperators();
    }

    [StorableConstructor]
    protected IntegerVectorPushProblem(bool deserializing)
      : base(deserializing) {
    }

    protected IntegerVectorPushProblem(IntegerVectorPushProblem original, Cloner cloner)
      : base(original, cloner) {
      InitEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitEvents();
    }

    private void InitEvents() {
      Config.EnabledExpressionsChanged += EnabledExpressionsChanged;
    }

    private void EnabledExpressionsChanged(object sender, EnabledExpressionsChangedEventArgs e) {
      Encoding.Bounds[0, 1] = Config.EnabledExpressions.Count;
      Encoding.BoundsParameter.Value[0, 1] = Config.EnabledExpressions.Count;
    }

    //protected override void OnReset() {
    //  base.OnReset();

    //  // clear pools and free reserved memory
    //  Pool.Clear();
    //  IndividualMapper.Clear();
    //  RandomPool.Clear();
    //  Config.Seed = 0;
    //}

    private void InitEncoding() {
      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = Config.EnabledExpressions.Count;
      Encoding.Length = Config.MaxPointsInProgram;
    }

    private void InitOperators() {

      var solutionCreator = Operators.OfType<PushSolutionCreator>().FirstOrDefault();

      if (solutionCreator == null) {
        solutionCreator = new PushSolutionCreator();
        Operators.Add(solutionCreator);
      }

      solutionCreator.ErcOptions = Config.ErcOptions;

      if (!Operators.OfType<PushExpressionFrequencyAnalyzer>().Any()) {
        Operators.Add(new PushExpressionFrequencyAnalyzer());
      }

      SolutionCreator = solutionCreator;
    }

    protected override PushProgram MapIndividual(Individual individual, IRandom random) {
      var program = individual.ToPushProgram(Config, random);

      return program;
    }

    public override double Evaluate(Individual individual, IRandom random) {
      // init seed of random pool
      //Interlocked.CompareExchange(ref RandomPool.Seed, random.Next(), 0);
      //Config.Seed = RandomPool.Seed;

      //var rand = RandomPool.ResetAndAllocate();
      var program = MapIndividual(individual, random);
      //RandomPool.Free(rand);

      //rand = RandomPool.ResetAndAllocate();
      var result = PushEvaluator.EvaluateTraining(Pool, program, random);
      //RandomPool.Free(rand);

      individual[CaseQualitiesScopeParameterName] = new DoubleArray(result.ExampleQualities);

      return result.AvgQuality;
    }
  }
}